""" Copyright start
  Copyright (C) 2008 - 2025 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import datetime
import logging
from datetime import timezone
from typing import Union

from xfutils.constants import DATE_FORMAT_STRING

logger = logging.getLogger(__name__)


def get_current_datetime():
    return datetime.datetime.now(datetime.timezone.utc)


def add_days_to_current_date(day):
    if day is None:
        raise ValueError("day cannot be None")
    return get_current_datetime() + datetime.timedelta(days=day)


def validate_and_return_date(date_string, expected_format):
    """
    Validates whether the given date string matches the expected format.
    If valid, returns a datetime object.

    Args:
        date_string (str): The date string to validate.
        expected_format (str): The expected date format (e.g., "%Y-%m-%d").

    Returns:
        datetime: The parsed datetime object if valid.
        None: If the input does not match the format.

    Raises:
        ValueError: If the date string is invalid or doesn't match the format.
    """
    try:
        # Parse the date string into a datetime object
        return datetime.datetime.strptime(date_string, expected_format)
    except ValueError as e:
        logger.error("Invalid date string: %s", date_string)
        # Return None or handle the error as needed
        return None


def calculate_date_difference(start_date: Union[str, datetime],
                              end_date: Union[str, datetime],
                              date_format: str = DATE_FORMAT_STRING) -> int:
    """
    Calculates the number of days between two dates.

    Args:
        start_date (str or datetime): The starting date as a string or a datetime object.
        end_date (str or datetime): The ending date as a string or a datetime object.
        date_format (str, optional): The format of the date string if a string is provided. Default is '%Y-%m-%d'.

    Returns:
        int: The number of days between the two dates.

    Raises:
        ValueError: If the provided date strings do not match the specified format.

    """
    if isinstance(start_date, str):
        try:
            # start_date = parser.isoparse(start_date)
            start_date = datetime.datetime.strptime(start_date, date_format)
            start_date = start_date.replace(tzinfo=timezone.utc)

        except ValueError:
            raise ValueError(f"Invalid format for start_date. Expected format: {date_format}")

    if isinstance(end_date, str):
        try:
            # end_date = parser.isoparse(end_date)
            end_date = datetime.datetime.strptime(end_date, date_format)
            end_date = end_date.replace(tzinfo=timezone.utc)

        except ValueError:
            raise ValueError(f"Invalid format for end_date. Expected format: {date_format}")

    delta = end_date - start_date
    return delta.days
