""" Copyright start
  Copyright (C) 2008 - 2025 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import base64
import hashlib
import logging
import secrets

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.padding import PKCS7
from django.conf import settings
from rest_framework import status

from xfutils.constants import APP_CONFIG_API_KEY, APP_CONFIG_API_SECRET, APP_CONFIG_API_FIXED_IV, AUTH_TYPE_BASIC, \
    UTF8
from xfutils.exceptions import XFException
from xfutils.import_export.error_message import xf_auth_45

logger = logging.getLogger(__name__)


def get_fixed_iv():
    return base64.b64decode(settings.ALL_CONFIG.get(APP_CONFIG_API_KEY, {}).get(APP_CONFIG_API_FIXED_IV))


def get_system_level_secret():
    """
    Get API key system secret which is required to encrypt tenant_id and API key
    Returns:
        str: Common API key system secret

    """
    return settings.ALL_CONFIG.get(APP_CONFIG_API_KEY, {}).get(APP_CONFIG_API_SECRET)


def generate_secret_key():
    """
    Generates a new Fernet secret key for symmetric encryption. This

    Returns:
        bytes: The generated secret key.
    """
    return base64.b64encode(secrets.token_bytes(32)).decode(UTF8)


def encrypt_data(key: str, plaintext: str) -> str:
    """
    Encrypts a given plaintext using AES-256 in CBC mode with a fixed IV to ensure deterministic encryption.

    Args:
        plaintext (str): The plaintext message to encrypt.
        key (bytes): The AES key in string.

    Returns:
        str: Base64-encoded ciphertext.
    """
    fixed_iv = get_fixed_iv()
    key = base64.b64decode(key)  # Convert string to bytes

    cipher = Cipher(algorithms.AES(key), modes.CBC(fixed_iv), backend=default_backend())

    # Pad plaintext to be multiple of block size (AES block size is 16 bytes)
    padder = PKCS7(128).padder()
    padded_text = padder.update(plaintext.encode(UTF8)) + padder.finalize()

    encryptor = cipher.encryptor()
    encrypted_bytes = encryptor.update(padded_text) + encryptor.finalize()

    # Convert encrypted bytes to Base64 for easier storage/sharing
    return base64.b64encode(encrypted_bytes).decode(UTF8)


def decrypt_data(key: str, ciphertext: str) -> str:
    """
    Decrypts a given ciphertext using AES-256 in CBC mode with the same fixed IV.

    Args:
        ciphertext (str): The Base64-encoded ciphertext to decrypt.
        key (bytes): The AES key in string.

    Returns:
        str: The original plaintext message.
    """
    fixed_iv = get_fixed_iv()
    key = base64.b64decode(key)
    cipher = Cipher(algorithms.AES(key), modes.CBC(fixed_iv), backend=default_backend())

    decryptor = cipher.decryptor()
    decrypted_padded = decryptor.update(base64.b64decode(ciphertext)) + decryptor.finalize()

    # Remove padding after decryption
    unpadder = PKCS7(128).unpadder()
    decrypted_text = unpadder.update(decrypted_padded) + unpadder.finalize()

    return decrypted_text.decode(UTF8)


def non_deterministic_encrypt_data(key, data):
    """
    Encrypts the given data using the Fernet symmetric encryption scheme.

    Args:
        key: The encryption key used for the encryption process
        data: The plain text data to be encrypted.

    Returns:
        str:  The encrypted data as a string.
    """
    fernet = Fernet(key)

    encrypted_message = fernet.encrypt(data.encode())  # Encrypt the message
    return encrypted_message.decode()


def non_deterministic_decrypt_data(key, encrypted_data):
    """
    Decrypts the given encrypted data using the Fernet symmetric encryption scheme.

    Args:
        key: The encryption key used for the Fernet encryption.
        encrypted_data: The encrypted data to decrypt.

    Returns:
        str: The decrypted plain text data.
    """
    fernet = Fernet(key.encode(UTF8))
    decrypted_message = fernet.decrypt(encrypted_data).decode()
    return decrypted_message


def hash_data(api_key):
    """
    Hashes the API key using SHA-256.

    Args:
        api_key (str): The API key to hash.

    Returns:
        str: The hashed API key.
    """

    api_key_bytes = api_key.encode(UTF8)
    hashed_key = hashlib.sha256(api_key_bytes).hexdigest()
    return hashed_key


def mask_data(data, visible_start=0, visible_end=4):
    """
    Mask an data, revealing only the first and last few characters.

    Args:
        data (str): data to be masked.
        visible_start (int): Number of characters to show at the start.
        visible_end (int): Number of characters to show at the end.

    Returns:
        str: The masked API key.
    """
    if len(data) <= visible_start + visible_end:
        return data

    return f"{data[:visible_start]}{'*' * (len(data) - visible_start - visible_end)}{data[-visible_end:]}"


def extract_basic_auth_credentials(auth_header: str) -> tuple[str, str]:
    """
    Extracts the username and password from a Basic Authentication header.

    Args:
        auth_header (str): The Authorization header value (e.g., "Basic dXNlcm5hbWU6cGFzc3dvcmQ=").

    Returns:
        tuple[str, str]: A tuple containing the username and password.

    Raises:
        ValueError: If the Authorization header is not properly formatted or invalid.
    """

    header_key = AUTH_TYPE_BASIC

    # Remove "Basic" word from provided header
    credential = auth_header[len(header_key):].strip() if auth_header.startswith(header_key) else auth_header

    try:
        decoded_credentials = decode_from_base64(credential)
    except Exception as e:
        raise XFException(status.HTTP_401_UNAUTHORIZED, xf_auth_45)

    # Split the decoded string into username and password
    if ":" not in decoded_credentials:
        raise XFException(status.HTTP_401_UNAUTHORIZED, xf_auth_45)

    username, password = decoded_credentials.split(":", 1)
    return username, password


def generate_random_key(length=32):
    """
    Generates a secure API key using the "secrets" module.

    Args:
        length (int, optional): The length of the API key to generate. Defaults to 32

    Returns:
        str:  A securely generated API key as a URL-safe string.

    Raises:
        ValueError: If the provided length is not a positive integer.
    """
    if length <= 0:
        raise ValueError("Length must be a positive integer.")

    return secrets.token_urlsafe(length)


def encode_to_base64(input_string: str) -> str:
    """
    Encodes a given string into Base64 format with exception handling.

    Parameters:
    input_string (str): The input string to be encoded.

    Returns:
    str: The Base64 encoded representation of the input string.
         Returns an error message if encoding fails.
    """
    try:
        encoded_bytes = base64.b64encode(input_string.encode(UTF8))
        return encoded_bytes.decode(UTF8)
    except Exception as e:
        return f"Error encoding to Base64: {str(e)}"


def decode_from_base64(encoded_string: str) -> str:
    """
    Decodes a Base64 encoded string back to its original form with exception handling.

    Parameters:
    encoded_string (str): The Base64 encoded string.

    Returns:
    str: The decoded original string.
         Returns an error message if decoding fails.
    """
    if not encoded_string or len(encoded_string.strip()) == 0:
        raise XFException(status.HTTP_401_UNAUTHORIZED, xf_auth_45)

    try:
        decoded_bytes = base64.b64decode(encoded_string)
        return decoded_bytes.decode(UTF8)
    except Exception as e:
        return f"Error decoding from Base64: {str(e)}"
