""" Copyright start
  Copyright (C) 2008 - 2025 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import ast
import json
import logging
import math

import jwt
import requests
from django_guid import get_guid
from gunicorn.glogging import Logger
from rest_framework import status
from rest_framework.exceptions import AuthenticationFailed

from xfutils.api_key_helper import ApiKeyHelper
from xfutils.constants import AUTH_TYPE_API_KEY, AUTH_TYPE_BASIC, AUTHORIZATION_KEY, AUTH_VERIFY_TOKEN
from xfutils.exceptions import XFException
from xfutils.import_export.error_message import xf_auth_34, NO_AUTH_HEADER
from xfutils.utilities.security_utils import extract_basic_auth_credentials

logger = logging.getLogger(__name__)


def normal_round(n):
    if n - math.floor(n) < 0.5:
        return math.floor(n)
    return math.ceil(n)


def api_call(url, method='GET', params=None, body='', headers=None, auth=None, verify=True,
             username='', password='', auth_config=None, timeout=None, auth_type='interservice', *args, **kwargs):
    if auth_config:
        auth = (auth_config['username'], auth_config['password'])
    elif username or password:
        auth = (username, password)

    verify = _convert_verify(verify)

    # build **args for requests call
    request_args = {
        'verify': verify,
    }
    if timeout:
        request_args['timeout'] = timeout
    if auth:
        request_args['auth'] = auth
    if params:
        request_args['params'] = params

    if not headers:
        headers = {}
    headers.update({'content-type': 'application/json'})

    if 'Correlation-ID' not in headers.keys():
        headers.update({'Correlation-ID': get_guid()})

    if kwargs.get('appliance_auth', False):
        if auth_type == 'interservice':
            token = generate_interservice_jwt_token(kwargs.get('all_config', {}), tenantid=kwargs.get('tenant', ''))
        elif auth_type == 'agent_appliance':
            token = generate_jwt_token(kwargs.get('all_config', {}).get('xf.instance.applianceUuid'),
                                       kwargs.get('all_config', {}).get('xf.instance.applianceKey'),
                                       tenantid=kwargs.get('tenant_id'), type='appliance')
        else:
            raise ValueError("Keyword argument auth_type must be either 'interservice' or 'agent_appliance' "
                             "when making an authenticated call with api_call()")
        headers.update({"Authorization": "Bearer {0}".format(token)})

    if headers:
        request_args['headers'] = headers

    # get rid of the body on GET/HEAD requests
    bodyless_methods = ['head', 'get']
    if method.lower() not in bodyless_methods:
        request_args['data'] = _convert_body(body)

    response = requests.request(method, url, **request_args)
    return maybe_json_or_raise(response, *args, **kwargs)


def generate_interservice_jwt_token(application_config, **kwargs):
    pvt_key_file = application_config.get('appliance', {}).get('private_key_filename')
    with open(pvt_key_file) as f:
        pvt_key = f.read()
    token = generate_jwt_token(application_config.get('appliance', {}).get('uuid'),
                               pvt_key, type='appliance', **kwargs)
    return token


def maybe_json_or_raise(response, *args, **kwargs):
    """
    Helper function for processing request responses

    Returns any json found in the response. Otherwise, it will extract the
    response as text, or, failing that, as bytes.

    :return: the response from the request
    :rtype: dict or str or bytes
    :raises: :class:`requests.HTTPError` if status code was 4xx or 5xx
    """
    if response.ok:
        try:
            logger.info('Processing request responses.')
            return response.json(strict=False)
        except Exception:
            if not kwargs.get('content'):
                return response.text or response.content
            else:
                return response.content or response.text
    else:
        msg = ''
        try:
            msg = response.json()
        except Exception:
            pass
        if not msg:
            msg = response.text

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # add any response content to the error message
            msg = f'{e} :: {msg}'
            raise requests.exceptions.HTTPError(msg, response=response)


def _convert_verify(verify):
    if verify and isinstance(verify, str):
        try:
            verify = ast.literal_eval(verify.title())
        except Exception as e:
            logger.warning(f'Str verification failed. {e}')
    if not isinstance(verify, bool):
        # just default to true
        return True
    return verify


def _convert_body(body):
    if body and isinstance(body, str):
        try:
            logger.info(f'converting body into json {body}')
            body = json.loads(body, strict=False)
        except Exception:
            logger.warning('Json conversion failed.')

    if body and not isinstance(body, str):
        body = json.dumps(body)
    return body


def generate_jwt_token(jti, private_key, **kwargs):
    auth_payload = kwargs
    auth_payload['jti'] = jti

    auth_header = {
        'alg': 'RS256'
    }
    token = jwt.encode(auth_payload, private_key, algorithm='RS256', headers=auth_header)
    try:
        token = token.decode('utf-8')
    except:
        pass
    return token


def get_claim_from_authorization(request):
    token = request.headers.get('Authorization')
    if not token:
        raise AuthenticationFailed(detail='Missing auth header in the request', code=status.HTTP_401_UNAUTHORIZED)

    credential = token.split(' ')
    _type = credential[0]
    _token = credential[1]

    if _type == AUTH_TYPE_API_KEY:
        return ApiKeyHelper.extract_information_from_key(_token)
    elif _type == AUTH_TYPE_BASIC:
        username, password = extract_basic_auth_credentials(_token)
        if username.lower() != AUTH_TYPE_API_KEY.lower():
            raise XFException(status.HTTP_401_UNAUTHORIZED, xf_auth_34)
        return ApiKeyHelper.extract_information_from_key(password)
    else:
        decode_token = jwt.decode(_token, algorithms=['RS256', 'HS256'], options={'verify_signature': False})
        return decode_token


def verify_token(request, auth_host):
    token = request.headers.get(AUTHORIZATION_KEY)
    if not token:
        raise AuthenticationFailed(detail=NO_AUTH_HEADER)

    url = AUTH_VERIFY_TOKEN.format(auth_host=auth_host)
    body = {'token': token}
    headers = {AUTHORIZATION_KEY: token}
    response = api_call(url, method='POST', body=body, headers=headers, verify=False)
    return response


#
# Gunicorn logger prints in below format
# [2024-04-03 03:58:58 +0000] [14] [DEBUG] POST /api/audit/query/audit-logs/
# Need to update its format to standard logging format, hence use GunicornCustomLogger.
#
class GunicornCustomLogger(Logger):
    def setup(self, cfg):
        super().setup(cfg)
        #
        # [None] [None] is correlation_id and tenantuuid. Since gunicorn is very first thing which does logging,
        # printing it as None. Note that, later, when django middleware runs, it prints correlation_id and tenantuuid.
        #
        formatter = logging.Formatter('[%(asctime)s] [%(process)d] [%(levelname)s] [None] [None] [%(name)s:%(lineno)d] %(message)s', datefmt='%d/%b/%Y %H:%M:%S %Z')
        if self.error_log.handlers:
            self.error_log.handlers[0].setFormatter(formatter)
        if self.access_log.handlers:
            self.access_log.handlers[0].setFormatter(formatter)
