""" Copyright start
  Copyright (C) 2008 - 2025 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import datetime
import logging
import os.path
import shutil
import urllib
import uuid

from django.conf import settings
from django.db import transaction
from django.http import HttpResponse
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, viewsets
from rest_framework.response import Response
from rest_framework.views import APIView
from xfutils.audit import publish_change_event
from xfutils.constants import DATE_FORMAT_STRING

from agents.constants import REMOTE_AGENT_OPERATIONS
from agents.exceptions import *
from agents.filtersets import AgentFilterSet
from agents.models import Agent
from agents.serializers import AgentSerializer, AgentReadOnlySerializer
from agents.utils.agent_count import has_room_for_agent
from agents.utils.agent_installer_utils import populate_installer_sh_file, populate_docker_agent_script
from agents.utils.appliance_utils import create_appliance_for_agent, delete_agent_appliance
from agents.utils.cert_utils import generate_keys_for_agent, get_pem_text
from agents.utils.rmq_mgmt import RMQ_Management
from connectors.core.constants import REMOTE_STATUS
from connectors.models import Connector
from connectors.serializers import ConnectorConfigurationSerializer
from connectors.utils import broadcast_to_remote, extract_audit_data_from_request
from contenthub.models import ContentHub
logger = logging.getLogger(__name__)
ENTITY_TYPES = {'agents': 'agents'}


# Create your views here.
class AgentViewSet(viewsets.ModelViewSet):
    queryset = Agent.objects.filter(isLocal=False).order_by('-createDate')
    serializer_class = AgentReadOnlySerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    http_method_names = ['get', 'post', 'put', 'delete']
    ordering_fields = ('name', 'healthStatus', 'enabled', 'createDate')
    filter_class = AgentFilterSet

    # Instead of transaction.atomic, set config to in-progress and then do the rmq status
    # if RMQ setup fails, set health status to failed
    @transaction.atomic
    def create(self, request, *args, **kwargs):
        self.serializer_class = AgentSerializer
        try:
            input_data = request.data
            if len(input_data['name']) > 100:
                msg = "Agent name should have a maximum of 100 characters"
                logger.error(msg)
                raise AgentCreationFailed(msg)
            if input_data.get('uuid'):
                input_data['uuid'] = uuid.UUID(input_data.get('uuid'))
            #@TODO: count will fail if we delete add agent
            try:
                has_room_for_agent()
            except AgentLimitReached as e:
                logger.error(e)
                raise e

            agent_id = str(uuid.uuid4()) if not input_data.get('agent_id') else input_data['agent_id']
            input_data['agentId'] = agent_id
            input_data['isLocal'] = False
            input_data['authType'] = 'certificate'
            # TODO: default to admin team once there is a good way
            # to differentiate between Teams objects
            input_data['owners'] = request.teams
            # Generate client certs and keys
            client_cert, client_key = generate_keys_for_agent(agent_id)
            input_data['clientCertificate'] = get_pem_text(client_cert)
            input_data['clientKey'] = get_pem_text(client_key, pem_type='private')
            input_data['healthStatus'] = {
                'status': 'awaiting',
                'message': 'Ready for Agent deployment'
            }
            # Create XF Appliance user for agent to authenticate as
            appliance = create_appliance_for_agent(agent_id, input_data['name'], request.tenantid, request.tenantuuid)
            input_data['applianceKey'] = appliance.get('appliance', {}).get('private_key')
            input_data['applianceUuid'] = appliance.get('appliance', {}).get('uuid')
            input_data['version'] = settings.RELEASE_VERSION
            agent_serializer = self.get_serializer(data=input_data)
            agent_serializer.is_valid(raise_exception=True)

            rmq_manager = RMQ_Management(agent_id)

            agent_obj = agent_serializer.save()
            headers = self.get_success_headers(agent_serializer.data)

            audit_data = AgentReadOnlySerializer(instance=agent_obj).data

            publish_change_event(request, 'integrations', 'create', ENTITY_TYPES['agents'],
                                 audit_data, agent_serializer.data['createDate'])
            return Response(audit_data, status=status.HTTP_201_CREATED, headers=headers)
        except CertificateGenerationFailed as e:
            logger.exception(e)
            raise e
        except ApplianceCreationFailed as e:
            raise e
        except Exception as e:
            logger.exception(e, exc_info=True)
            raise AgentCreationFailed(e.args)

    @transaction.atomic
    def destroy(self, request, *args, **kwargs):
        self.serializer_class = AgentSerializer
        try:
            obj = self.get_object()
            deleted_at = datetime.datetime.now().strftime(DATE_FORMAT_STRING)
            serializer_data = self.get_serializer(obj, context={'request': request}).data
            rmq_manager = RMQ_Management(obj.agentId, operation='spindown')
            obj.delete()
            delete_agent_appliance(obj.agentId, str(obj.applianceUuid), request.tenantid, request.tenantuuid)
            publish_change_event(request, 'integrations', 'delete', ENTITY_TYPES['agents'],
                                 serializer_data, deleted_at)
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Exception as e:
            logger.exception(e, exc_info=True)
            raise AgentDeleteFailed(e.args)

    def update(self, request, *args, **kwargs):
        self.serializer_class = AgentSerializer
        try:
            input_data = request.data
            obj = self.get_object()
            was_active = obj.active
            serializer = self.get_serializer(obj, data=input_data, partial=True)
            # Validate user-submitted data before continuing
            serializer.is_valid(raise_exception=True)
            
            require_health_check = False
            
            if was_active and not input_data.get('active', was_active):
                input_data['healthStatus'] = {
                    'status': 'inactive',
                    'message': 'Agent is inactive'
                }
                
            elif not was_active and input_data.get('active', was_active):
                rmq_handler = RMQ_Management(obj.agentId)
                input_data['healthStatus'] = {
                    'status': 'disconnected',
                    'message': 'No active Agent connections'
                }
                require_health_check = True
            # Need a new serializer since the validated one can no longer be modified
            serializer = self.get_serializer(obj, data=input_data, partial=True)
            serializer.is_valid()
            obj = serializer.save()

            if require_health_check:
                message_json = {
                    "data": {"agent": obj.agentId},
                    "sourceId": settings.SELF_ID,
                    "destinationId": obj.agentId,
                    "action": REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'),
                    "method": "POST",
                    "tenant": request.tenantid
                }
                message_json['data']['request_data'] = extract_audit_data_from_request(request)
                logger.error(f"Caling broadcast function to publish healthcheck message")
                broadcast_to_remote(message_json, obj.agentId, REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'), 'receiveConnectorInstructionRequest', 'POST')

            headers = self.get_success_headers(serializer.data)
            if getattr(obj, '_prefetched_objects_cache', None):
                # If 'prefetch_related' has been applied to a queryset, we need to
                # forcibly invalidate the prefetch cache on the instance.
                obj._prefetched_objects_cache = {}
            audit_data = AgentReadOnlySerializer(instance=obj).data
            publish_change_event(request, 'integrations', 'update', ENTITY_TYPES['agents'],
                                 audit_data, serializer.data['modifyDate'])
            return Response(audit_data, status=status.HTTP_200_OK, headers=headers)
        except Exception as e:
            logger.exception(e, exc_info=True)
            raise AgentUpdateFailed(e.args)


    def bulk_delete(self, request, *args, **kwargs):
        agent_ids = request.data.get('agents', [])
        agent_objs = Agent.objects.filter(agentId__in=agent_ids)
        spundown_agents = []
        appliances = []
        for agent in agent_objs:
            try:
                rmq = RMQ_Management(agent.agentId, operation='spindown')
                spundown_agents.append(agent.agentId)
                appliances.append(str(agent.applianceUuid))
            except Exception as e:
                message = f"Could not delete Agent {agent} :: ERROR: {e.message}"
                logger.exception(message)
                if spundown_agents:
                    spundown_agents = Agent.objects.filter(agentId__in=spundown_agents)
                    spundown_agents.delete()
                raise AgentDeleteFailed(message)
        agent_objs.delete()
        delete_agent_appliance("", appliances, request.tenantid, request.tenantuuid)
        return Response({"message": "success"}, status=status.HTTP_200_OK)
    
    def trigger_healthcheck(self, request, *args, **kwargs):
        try:
            agent_id = request.data.get('agent', None)
            if not agent_id:
                agent_id = request.GET.get('agent', None)

            agent_obj = Agent.objects.get(agentId=agent_id)

            # First use RMQ management to see if there are any consumers on the agent vhost
            rmq_connections = RMQ_Management(agent_id=agent_id, operation='connections').connections
            if len(rmq_connections) == 0:
                if agent_obj.healthStatus['status'] != 'awaiting':
                    agent_obj.healthStatus = {
                        'status': 'disconnected',
                        'message': 'No active Agent connections'
                    }
                    agent_obj.save()
                return Response(agent_obj.healthStatus, status=status.HTTP_200_OK)
            rmq_data = {
                'agent': agent_id,
                'token': request.headers.get('Authorization'),
                'request_data': extract_audit_data_from_request(request)
            }
            message_json = {
                "data": rmq_data,
                "sourceId": settings.SELF_ID,
                "destinationId": agent_obj.agentId,
                "action": REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'),
                "method": "POST",
                "token": rmq_data.get('token'),
                "tenant": request.tenantid
            }
            agent_obj.healthStatus = {
                    'status': 'in progress',
                    'message': 'Agent health check in progress'
            }
            agent_obj.save()
            broadcast_to_remote(message_json, agent_id, REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'), 'receiveConnectorInstructionRequest', 'POST')
            return Response(agent_obj.healthStatus, status=status.HTTP_200_OK)
        except Agent.DoesNotExist:
            error_message = 'Agent with id {0} not found'.format(agent_id)
            logger.error(error_message)
            return Response({"message": error_message}, status=status.HTTP_400_BAD_REQUEST)
        except Exception as e:
            logger.error(str(e))
            agent_obj.healthStatus = {
                'status': REMOTE_STATUS.get('disconnected'),
                'message': str(e)
            }
            agent_obj.save()
            return Response(agent_obj.healthStatus, status=status.HTTP_200_OK)

    
    def health_status(self, request, *args, **kwargs):
        agent_id = request.GET.get('agent', None)
        agent_obj = Agent.objects.get(agentId=agent_id)
        return Response(agent_obj.healthStatus, status=status.HTTP_200_OK)
    

class AgentInstallerView(APIView):
    http_method_names = ['post']

    def post(self, request, *args, **kwargs):
        request_data = request.data
        agent = request_data.get('agent')

        try:
            agent_obj = Agent.objects.get(agentId=agent)
        except Agent.DoesNotExist:
            return Response({'message': f'Agent {agent} does not exist.'}, status=status.HTTP_404_NOT_FOUND)

        hostname = request.get_host()

        # If the "include_last_known_configurations" attribute is True, then ignore any supplied connectors.
        # Instead, get all the connectors installed on the Agent, and their configurations, and use that as the connectors
        # object for the request
        if request.data.get('include_last_known_configurations', None):
            agent_connectors = Connector.objects.filter(agent=agent_obj, installed=True)
            connectors = []
            for c in agent_connectors:
                #agent_connector_configurations = Configuration.objects.filter(agent=agent_obj, connector__in=agent_connectors)
                configuration_data = ConnectorConfigurationSerializer(c.configuration.all(), many=True).data
                for conf in configuration_data:
                    conf.pop('id', None)
                    conf.pop('connector', None)
                    conf.pop('health_status', None)
                connectors.append({
                    'name': c.name,
                    'version': c.version,
                    'configurations': configuration_data
                })
        else:
            connectors = request.data.get('connectors', [])

        connector_wo_version = []
        default_connector_names = ['cyops_utilities']
        for connector in connectors:
            # find out connectors without version
            if not connector.get("version"):
                 connector_wo_version.append(connector.get("name"))
            # find out which default connector is included or not
            if connector.get("name") in default_connector_names:
                default_connector_names.remove(connector.get("name"))

        if default_connector_names:
            # get all default connectors which are not include in connectors json
            # append them in connectors json
            default_connectors = Connector.objects.filter(name__in=default_connector_names, agent_id=settings.SELF_ID,
                                                            installed=True)
            for default_connector in default_connectors:
                connectors.append({
                    'name': default_connector.name,
                    'version': default_connector.version
                })

        if  connector_wo_version:
            # get all connector which version is not mentioned
            # update the version in connectors json
            latest_connectors = ContentHub.objects.filter(name__in= connector_wo_version)

            for latest_connector in  latest_connectors:
                for connector in connectors:
                    if connector.get("name") == latest_connector.name:
                        connector["version"] = latest_connector.version
                        break

        install_method = request_data.get('deployment', 'docker')
        if install_method == 'docker':
            file_func = populate_docker_agent_script
            filename = f'deploy-docker-agent-{agent_obj.agentId}.sh'
        elif install_method == 'rpm':
            file_func = populate_installer_sh_file
            filename = 'install-agent-{agent_obj.agentId}.sh'
        filename = urllib.parse.quote(filename)
        try:
            install_file = file_func(agent_obj, tenantid=request.tenantid, hostname=hostname, connectors=connectors)
            with open(install_file, 'rb') as instFile:
                content = instFile.read()
            response = HttpResponse(content)
            response['content_type'] = 'application/text'
            response['Content-Disposition'] = f'attachment; filename={filename}'
            return response
        except Exception as e:
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        finally:
            if os.path.isfile(install_file):
                shutil.rmtree(install_file, ignore_errors=True)

        # if install_method == 'docker':
        #     docker_file = populate_docker_compose_file(agent_obj, tenantid=request.tenantid)

        #     try:
        #         with open(docker_file, 'rb') as shfile:
        #             content = shfile.read()
        #         response = HttpResponse(content)
        #         response['content_type'] = 'application/text'
        #         response['Content-Disposition'] = 'attachment; filename=docker-compose-agent.yaml'
        #         shutil.rmtree(docker_file, ignore_errors=True)
        #         return response
        #     except Exception as e:
        #         return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        # elif install_method == 'rpm':
        #     sh_file = populate_installer_sh_file(agent_obj, tenantid=request.tenantid)