""" Copyright start
  Copyright (C) 2008 - 2025 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import logging
import uuid

from django.db.models import F
from django_tenants.utils import tenant_context
from rest_framework.request import HttpRequest, Request
from taggit.models import Tag

from xfutils.import_export.utils import XFSynchronousPorterActions

from agents.models import Agent
from connectors.models import Connector, Configuration
from connectors.serializers import ConnectorListSerializer, ConnectorConfigurationSerializer
from connectors.views import ConnectorInstall
from constants.error_message import xf_integration_57
from tenant.models import Tenant

logger = logging.getLogger(__name__)

MODEL_MAPPER = {
    'connectors': {
        'model': Connector,
        'serializer': ConnectorListSerializer,
        'exclude_fields': ['status', 'remote_status', 'rpm_installed', 'id', 'category',
                           'active', 'system', 'icon_small', 'iconLarge', 'description',
                           'install_result', 'ingestion_supported', 'tags', 'development',
                           'rpm_full_name', 'help_online', 'publisher', 'type', 'modified',
                           'agent']
    }
}


class SynchronousPorterActions(XFSynchronousPorterActions):

    def run_export(*args, **kwargs):
        data = {}
        serialized_request_attrs = kwargs.get('serialized_request_attrs', {})
        tenant_id = serialized_request_attrs.get('tenantid')
        tenant = Tenant.objects.get(tenant_id=tenant_id)
        with tenant_context(tenant):
            agent_id = Agent.objects.get(isLocal=True).agentId
            for model_name, selections in kwargs.get('selections').items():
                if model_name in MODEL_MAPPER:
                    unique_list = [selection.get('value') for selection in selections]
                    connectors_with_configs = [selection.get('value') for selection in selections if
                                               selection.get('configurations', True)]
                    if model_name == 'connectors':
                        objects = Connector.objects.filter(name__in=unique_list, agent=agent_id)
                    else:
                        objects = MODEL_MAPPER[model_name]['model'].objects.filter(uuid__in=unique_list)
                    serialized_objects = MODEL_MAPPER[model_name]['serializer'](objects, many=True).data
                    if model_name == 'connectors':
                        for connector_object in serialized_objects:
                            if connector_object.get('name', '') in connectors_with_configs:
                                config_objects = Configuration.objects.filter(connector_id=connector_object.get('id'))
                                config_data = ConnectorConfigurationSerializer(config_objects, many=True).data
                                for config in config_data:
                                    config.pop('id', None)
                                    config.pop('remote_status', None)
                                    config.pop('health_status', None)
                                    config.pop('agent', None)
                                    config.pop('connector', None)
                                    config.pop('teams', None)
                                connector_object['configuration'] = config_data
                            else:
                                connector_object.pop('configuration', None)
                            # deferring the filter query doesn't seem to work, so we'll do extra labor here
                            for field_to_exclude in MODEL_MAPPER['connectors']['exclude_fields']:
                                connector_object.pop(field_to_exclude, None)
                    data[model_name] = serialized_objects
        return data

    def build_import_options(*args, **kwargs):
        all_options = {}
        tenant_uuid = kwargs.get('tenant')
        tenant = Tenant.objects.get(uuid=tenant_uuid)
        with tenant_context(tenant):
            agent_id = Agent.objects.get(isLocal=True).agentId
            for model_name, selections in kwargs.get('file_data').items():
                sub_options = []
                if model_name in MODEL_MAPPER:
                    # connectors exist has the (name, agent) uniqueness constraint and not uuid
                    if model_name == 'connectors':  # preemptive check here in case other integration models are included in the future
                        found_unique_existing = {}
                        for item in selections:
                            unique_existing = MODEL_MAPPER[model_name]['model'].objects.filter(name=item['name'],
                                                                                               agent=agent_id)
                            if unique_existing:
                                found_unique_existing[item['name']] = (item['label'], item['version'], item.get('config_count', 0))
                        for name, details in found_unique_existing.items():
                            sub_options.append({
                                'value': name,
                                'label': details[0],
                                'version': details[1],
                                'configCount': details[2],
                                'exists': True,
                                'include': True
                            })
                        for selection in selections:
                            if selection.get('name') not in found_unique_existing:
                                sub_options.append({
                                    'value': selection.get('name'),
                                    'label': selection.get('label'),
                                    'version': selection.get('version'),
                                    'configCount': selection.get('config_count', 0),
                                    'exists': False,
                                    'include': True
                                })
                    else:
                        selection_uuids = [item['uuid'] for item in selections]
                        found_existing = MODEL_MAPPER[model_name]['model'].objects.filter(uuid__in=selection_uuids)
                        existing_uuids = [str(entry.uuid) for entry in found_existing]
                        for existing_entry in found_existing:
                            entry_name = existing_entry.name
                            sub_options.append({
                                'value': str(existing_entry.uuid),
                                'label': entry_name,
                                'exists': True,
                                'include': True
                            })
                        for selection in selections:
                            if selection.get('uuid') not in existing_uuids:
                                sub_options.append({
                                    'value': selection.get('uuid'),
                                    'label': selection.get('name'),
                                    'exists': False,
                                    'include': True
                                })
                all_options[model_name] = sub_options
        return all_options

    def run_import(*args, **kwargs):
        # Return object will be in this format
        # {
        #     "connector":{
        #         "status": "",
        #         "message": ""
        #     }
        # }
        options = kwargs.get('options')
        file_data = kwargs.get('file_data')
        serialized_request_attrs = kwargs.get('serialized_request_attrs', {})
        tenant = Tenant.objects.get(tenant_id=serialized_request_attrs.get('tenantid'))
        all_status = ''
        message = 'All connectors installed successfully'
        failed_connectors = []
        with tenant_context(tenant):
            agent_id = Agent.objects.get(isLocal=True).agentId
            for model_name, selections in options.items():
                if model_name in MODEL_MAPPER:
                    included_records = []
                    bulk_records = []
                    bulk_configs = []
                    update_fields = []
                    to_import = [selection for selection in selections if selection.get('include')]
                    for option in to_import:
                        if model_name == 'connectors':
                            file_record = [file for file in file_data[model_name] if file.get('name') == option.get('value')][0]
                            file_record['exists'] = option.get('exists', False)
                        else:
                            file_record = [file for file in file_data[model_name] if file.get('uuid') == option.get('value')][0]
                        file_record.pop('tenant', None)
                        file_record.pop('users', None)
                        connector_configs = file_record.pop('configuration', None)
                        if connector_configs:
                            bulk_configs.append({'name': file_record.get('name', ''),
                                                 'agent': agent_id,
                                                 'configuration': connector_configs})
                        included_records.append(file_record)
                        if not update_fields:
                            update_fields = list(file_record.keys())
                            try:
                                update_fields.remove('uuid')
                            except ValueError:
                                pass
                    if model_name == 'connectors':
                        integration_messages = {}
                        for connector in included_records:
                            if not connector.get('exists'):
                                request = Request(HttpRequest())
                                serialized_request_attrs = kwargs.get('serialized_request_attrs', {})
                                for attribute, value in serialized_request_attrs.items():
                                    if attribute == 'user_uuid':
                                        setattr(request, 'useruuid', value)
                                    else:
                                        setattr(request, attribute, value)
                                request.data.update({
                                    'name': connector.get('name'),
                                    'agent': agent_id
                                })
                                if model_name not in integration_messages:
                                    integration_messages[model_name] = {}
                                try:
                                    ConnectorInstall.install(ConnectorInstall, request)
                                    # If status need per connector we can use this format
                                    # integration_messages[model_name].update({
                                    #     connector.get('name'): {'status': 'success'}
                                    # })
                                    all_status = 'partial-success' if all_status == 'failed' else 'success'
                                except Exception as e:
                                    # connector install failed
                                    error_message = xf_integration_57.format(connector.get('name'))
                                    logger.error("{0} Error : {1}".format(error_message, str(e)))
                                    # If status need per connector we can use this format
                                    # integration_messages[model_name].update({
                                    #     connector.get('name'): {'status': 'failed', 'message': error_message}
                                    # })
                                    failed_connectors.append(connector.get('name'))
                                    all_status = 'partial-success' if all_status == 'success' else 'failed'
                        if bulk_configs:
                            for connector in bulk_configs:
                                tag_dict = {}
                                all_tags = []
                                tag_objects = []
                                config_objects = []
                                for config in connector.get('configuration', []):
                                    config_object = Configuration(**config)
                                    tag_dict[config.get('config_id', '')] = config.get('tags', [])
                                    all_tags.extend(config.get('tags', []))
                                    try:
                                        connector_object = Connector.objects.get(name=connector.get('name', ''),
                                                                                 agent=connector.get('agent', agent_id))
                                        config_object.connector = connector_object
                                        config_object.agent_id = agent_id
                                        config_objects.append(config_object)
                                    except Exception as e:
                                        # should never throw exception assuming the connector install above was successful
                                        logger.error(e)
                                try:
                                    Configuration.objects.bulk_create(config_objects,
                                                                      update_conflicts=True,
                                                                      update_fields=['config', 'default'],
                                                                      unique_fields=['config_id'])
                                    config_count = len(Configuration.objects.filter(connector=connector_object,
                                                                                    agent_id=agent_id))
                                    Connector.objects.filter(
                                        name=connector.get('name', ''),
                                        agent=connector.get('agent', agent_id)).update(config_count=config_count)
                                except Exception:
                                    # name conflict, append index to name and create new
                                    for config in config_objects:
                                        configuration_index = 1
                                        old_config_name = config.name
                                        while True:
                                            config.name = f'{old_config_name} ({configuration_index})'
                                            try:
                                                config.config_id = uuid.uuid4()
                                                config.save()
                                                Connector.objects.filter(name=config.connector.name,
                                                                         agent=agent_id).update(config_count=F('config_count')+1)
                                                break
                                            except Exception:
                                                configuration_index += 1
                                                continue
                            # import config tags
                            if all_tags:
                                for tag in all_tags:
                                    tag_objects.append(Tag(name=tag, slug=tag))
                                Tag.objects.bulk_create(tag_objects, ignore_conflicts=True)

                                tags_from_db = Tag.objects.all()
                                configs_from_db = Configuration.objects.filter(config_id__in=list(tag_dict.keys()))
                                for config in configs_from_db:
                                    related_tags = tags_from_db.filter(name__in=tag_dict[str(config.config_id)])
                                    config.tags.add(*related_tags)
                    else:
                        integration_messages = {model_name: []}
                        try:
                            MODEL_MAPPER[model_name]['model'].objects.bulk_create(bulk_records,
                                                                                  update_conflicts=True,
                                                                                  update_fields=update_fields,
                                                                                  unique_fields=['uuid'])

                            all_status = 'success'

                        except Exception as e:
                            all_status = 'failed'
                            error_message = "Error while bulk creating {0} ERROR: {1}".format(model_name, str(e))
                            logger.error(error_message)
                            message = error_message


                    integration_messages[model_name] = {
                        'status': all_status,
                        'message': '['+','.join(failed_connectors) + '] connectors failed' if failed_connectors else message
                    }
                    return integration_messages
